unless window.coffeetable then window.coffeetable = {}
app = window.coffeetable

app.beats = {

	reduce_beats: (beat_indices, type) ->

		###

		Applies either union or intersection (as determined by `type`) to the progressions
		collected in the array `beat_indices`, returning a single, sorted set of integers.

		###

		if beat_indices 
			beat_indices = _.compact(beat_indices)
			unless beat_indices.length == 0
				if type == 'and'
					beats = _.intersection.apply(null, beat_indices)
				else 
					beats = _.union.apply(null, beat_indices)
				_.sortBy(beats, (num) -> num)

	read_generator: (gen_el, measures, resolution) ->

		###

		Produces an arithmetic progression, given a generator element, the measure count,
		and the resolution. Returns an array of integers: 
		[offset + 1...measures * resolution] by distance.

		###

		if $(gen_el).hasClass('active')
			values = []
			$(gen_el).find('input').each((i) -> values.push(parseInt($(this).val())))
			[step, offset] = values
			_.range(offset + 1, (measures * resolution) + 1, step)
		else
			null

	read_row: (row_el, measures, resolution) ->

		###

		Calls `read_generator` on each of the generators in this row, and reduces them
		with `reduce_beats`; if there's a product it's returned, otherwise `null`.

		###

		type = $('.toggle-box .selected', row_el).text()
		beats = (app.beats.read_generator(gen_el, 
			measures, resolution) for gen_el in $('.generator-box', row_el))
		row_beats = app.beats.reduce_beats(beats, type)
		if row_beats then row_beats else null

	get_active_beats: ->

		###

		Calls `read_row` on all rhythm-rows; reduces the product with `reduce_beats`;
		returns the reduction if any or null.

		###

		measures = app.parameters.measures()
		resolution = app.parameters.resolution()
		beats = (app.beats.read_row(row_el, measures, 
			resolution) for row_el in $('.rhythm-row'))
		active_beats = app.beats.reduce_beats(beats, 'or')
		if active_beats then active_beats else null

	get_measure_number: (pos, res, msr) ->

		###

		Employs a touch of recursion to figure out what measure a given beat should be in.

		###

		unless msr then msr = 1
		if (pos - (msr * res)) > 0
			app.beats.get_measure_number(pos, res, msr + 1)
		else
			msr

	adjust_position_for_measure: (pos, res) ->

		###

		Takes a raw beat index (e.g. 17 at a `res` of 16) and returns a measure, position
		pair (e.g. [2, 1] using the example above).

		###

		msr = app.beats.get_measure_number(pos, res)
		[msr, (res - ((res * msr) - pos))]

	get_adjusted_beat_positions: (res, active_beats) ->

		###
	
		Maps `adjust_position_for_measure` across all active beats, returning a 2d array:
		`[[1, 1], [1, 2] ...]`

		###

		(app.beats.adjust_position_for_measure(pos, res) for pos in active_beats)


	x_offset: (width, beat) -> # partial
		(x) -> x + (width * (beat - 1))

	y_offset: (height, margin, msr) -> # partial
		(y) -> y + (height + margin) * (msr - 1)

	amp_env: (w, h, margin, attack, decay, sust, rel) => # partial

		###

		Returns a function that computes the coordinates for 3 quadratic curves and a line 
		(representing an amplitude envelope) from the start to the end of a given 
		measure / beat.

		###

		(msr, beat) =>
			[a, d, s, r] = _.map([attack, decay, sust, rel], (v) -> v * 0.01)
			x_off = app.beats.x_offset(w, beat)
			y_off = app.beats.y_offset(h, margin, msr)
			attack_length = a * w
			decay_length = d * (w - attack_length)
			release_length = r * (w - (attack_length + decay_length))
			base_x = x_off(0)
			attack_x = x_off(attack_length)
			decay_x = x_off(attack_length + decay_length)
			release_x = x_off(attack_length + decay_length + release_length)
			rightmost_x = x_off(w)
			base_y = y_off(h)
			peak_y = y_off(margin)
			sus_y = y_off(margin + ((h - margin) - ((h - margin) * s)))
			ctx_1_x = x_off(attack_length / 2)
			ctx_2_x = x_off(attack_length + (decay_length / 2))
			ctx_3_x = x_off(attack_length + decay_length + (release_length / 2))
			[
				["quadraticCurveTo", ctx_1_x, peak_y, attack_x, peak_y],
				["quadraticCurveTo", ctx_2_x, sus_y, decay_x, sus_y],
				["quadraticCurveTo", ctx_3_x, sus_y, release_x, base_y],
				["lineTo", rightmost_x, base_y]
			]

	rest_beat: (w, h, margin) -> # partial

		###

		Returns a function that computes the start and end coordinates of a flat line,
		representing a rest at the given measure / beat.

		###

		(msr, beat) ->
			x_off = app.beats.x_offset(w, beat)
			y_off = app.beats.y_offset(h, margin, msr)
			[["lineTo", x_off(w), y_off(h)]]

	get_beat_matrix: (res, msrs, active_beats) ->

		###

		Produces a matrix `msrs` x `res`, where each entry is an array with 
		information about location and whether the note is active or a rest.

		###

		beat_matrix = []
		_.each(_.range(1, msrs + 1), (m) =>
			i = m - 1
			unless beat_matrix[i] then beat_matrix[i] = []
			_.each(_.range(1, res + 1), (b) =>
				if app.util.compare_nested_array(active_beats, m, b)
					this_beat = [m, b, 'active']
				else
					this_beat = ([m, b, 'rest'])
				beat_matrix[i].push(this_beat)))
		beat_matrix

	draw_measure: (w, h, margin, a, d, s, r) -> # partial

		###

		Returns a function that takes one row from the beat_matrix (`msr`) and returns
		a set of drawing instructions for that measure.

		###

		(msr) ->
			active_b = app.beats.amp_env(w, h, margin, a, d, s, r)
			rest_b = app.beats.rest_beat(w, h, margin)
			instructions = [['beginPath'], ['moveTo', 0, (h + (h + margin) * (msr[0][0] - 1))]]
			end = [['closePath'], ['stroke'], ['fill']]
			
			_.each(msr, (beat) ->
				[m, b, type] = beat
				instrs = if type == 'active' then active_b(m, b) else rest_b(m, b)
				_.each(instrs, (instr) -> 
					instructions.push(instr)))
			
			_.each(end, (instr) -> 
				instructions.push(instr))
			
			instructions

	get_instrs: (m, res, a, d, s, r, active_beats) ->

		###

		Produces a beat matrix, and maps `draw_measure` against it, returning a full
		set of beatmap drawing instructions. Height and margin defaults are set here.

		###

		beat_info = app.beats.get_adjusted_beat_positions(res, active_beats)
		w = 1024 / res
		h = 100
		margin = 10
		drawer = app.beats.draw_measure(w, h, margin, a, d, s, r)
		beat_matrix = app.beats.get_beat_matrix(res, m, beat_info)
		(drawer(msr) for msr in beat_matrix)
}