(function() {
  var app,
    _this = this;

  if (!window.coffeetable) window.coffeetable = {};

  app = window.coffeetable;

  app.beats = {
    reduce_beats: function(beat_indices, type) {
      /*
      
      		Applies either union or intersection (as determined by `type`) to the progressions
      		collected in the array `beat_indices`, returning a single, sorted set of integers.
      */
      var beats;
      if (beat_indices) {
        beat_indices = _.compact(beat_indices);
        if (beat_indices.length !== 0) {
          if (type === 'and') {
            beats = _.intersection.apply(null, beat_indices);
          } else {
            beats = _.union.apply(null, beat_indices);
          }
          return _.sortBy(beats, function(num) {
            return num;
          });
        }
      }
    },
    read_generator: function(gen_el, measures, resolution) {
      /*
      
      		Produces an arithmetic progression, given a generator element, the measure count,
      		and the resolution. Returns an array of integers: 
      		[offset + 1...measures * resolution] by distance.
      */
      var offset, step, values;
      if ($(gen_el).hasClass('active')) {
        values = [];
        $(gen_el).find('input').each(function(i) {
          return values.push(parseInt($(this).val()));
        });
        step = values[0], offset = values[1];
        return _.range(offset + 1, (measures * resolution) + 1, step);
      } else {
        return null;
      }
    },
    read_row: function(row_el, measures, resolution) {
      /*
      
      		Calls `read_generator` on each of the generators in this row, and reduces them
      		with `reduce_beats`; if there's a product it's returned, otherwise `null`.
      */
      var beats, gen_el, row_beats, type;
      type = $('.toggle-box .selected', row_el).text();
      beats = (function() {
        var _i, _len, _ref, _results;
        _ref = $('.generator-box', row_el);
        _results = [];
        for (_i = 0, _len = _ref.length; _i < _len; _i++) {
          gen_el = _ref[_i];
          _results.push(app.beats.read_generator(gen_el, measures, resolution));
        }
        return _results;
      })();
      row_beats = app.beats.reduce_beats(beats, type);
      if (row_beats) {
        return row_beats;
      } else {
        return null;
      }
    },
    get_active_beats: function() {
      /*
      
      		Calls `read_row` on all rhythm-rows; reduces the product with `reduce_beats`;
      		returns the reduction if any or null.
      */
      var active_beats, beats, measures, resolution, row_el;
      measures = app.parameters.measures();
      resolution = app.parameters.resolution();
      beats = (function() {
        var _i, _len, _ref, _results;
        _ref = $('.rhythm-row');
        _results = [];
        for (_i = 0, _len = _ref.length; _i < _len; _i++) {
          row_el = _ref[_i];
          _results.push(app.beats.read_row(row_el, measures, resolution));
        }
        return _results;
      })();
      active_beats = app.beats.reduce_beats(beats, 'or');
      if (active_beats) {
        return active_beats;
      } else {
        return null;
      }
    },
    get_measure_number: function(pos, res, msr) {
      /*
      
      		Employs a touch of recursion to figure out what measure a given beat should be in.
      */      if (!msr) msr = 1;
      if ((pos - (msr * res)) > 0) {
        return app.beats.get_measure_number(pos, res, msr + 1);
      } else {
        return msr;
      }
    },
    adjust_position_for_measure: function(pos, res) {
      /*
      
      		Takes a raw beat index (e.g. 17 at a `res` of 16) and returns a measure, position
      		pair (e.g. [2, 1] using the example above).
      */
      var msr;
      msr = app.beats.get_measure_number(pos, res);
      return [msr, res - ((res * msr) - pos)];
    },
    get_adjusted_beat_positions: function(res, active_beats) {
      /*
      	
      		Maps `adjust_position_for_measure` across all active beats, returning a 2d array:
      		`[[1, 1], [1, 2] ...]`
      */
      var pos, _i, _len, _results;
      _results = [];
      for (_i = 0, _len = active_beats.length; _i < _len; _i++) {
        pos = active_beats[_i];
        _results.push(app.beats.adjust_position_for_measure(pos, res));
      }
      return _results;
    },
    x_offset: function(width, beat) {
      return function(x) {
        return x + (width * (beat - 1));
      };
    },
    y_offset: function(height, margin, msr) {
      return function(y) {
        return y + (height + margin) * (msr - 1);
      };
    },
    amp_env: function(w, h, margin, attack, decay, sust, rel) {
      /*
      
      		Returns a function that computes the coordinates for 3 quadratic curves and a line 
      		(representing an amplitude envelope) from the start to the end of a given 
      		measure / beat.
      */      return function(msr, beat) {
        var a, attack_length, attack_x, base_x, base_y, ctx_1_x, ctx_2_x, ctx_3_x, d, decay_length, decay_x, peak_y, r, release_length, release_x, rightmost_x, s, sus_y, x_off, y_off, _ref;
        _ref = _.map([attack, decay, sust, rel], function(v) {
          return v * 0.01;
        }), a = _ref[0], d = _ref[1], s = _ref[2], r = _ref[3];
        x_off = app.beats.x_offset(w, beat);
        y_off = app.beats.y_offset(h, margin, msr);
        attack_length = a * w;
        decay_length = d * (w - attack_length);
        release_length = r * (w - (attack_length + decay_length));
        base_x = x_off(0);
        attack_x = x_off(attack_length);
        decay_x = x_off(attack_length + decay_length);
        release_x = x_off(attack_length + decay_length + release_length);
        rightmost_x = x_off(w);
        base_y = y_off(h);
        peak_y = y_off(margin);
        sus_y = y_off(margin + ((h - margin) - ((h - margin) * s)));
        ctx_1_x = x_off(attack_length / 2);
        ctx_2_x = x_off(attack_length + (decay_length / 2));
        ctx_3_x = x_off(attack_length + decay_length + (release_length / 2));
        return [["quadraticCurveTo", ctx_1_x, peak_y, attack_x, peak_y], ["quadraticCurveTo", ctx_2_x, sus_y, decay_x, sus_y], ["quadraticCurveTo", ctx_3_x, sus_y, release_x, base_y], ["lineTo", rightmost_x, base_y]];
      };
    },
    rest_beat: function(w, h, margin) {
      /*
      
      		Returns a function that computes the start and end coordinates of a flat line,
      		representing a rest at the given measure / beat.
      */      return function(msr, beat) {
        var x_off, y_off;
        x_off = app.beats.x_offset(w, beat);
        y_off = app.beats.y_offset(h, margin, msr);
        return [["lineTo", x_off(w), y_off(h)]];
      };
    },
    get_beat_matrix: function(res, msrs, active_beats) {
      /*
      
      		Produces a matrix `msrs` x `res`, where each entry is an array with 
      		information about location and whether the note is active or a rest.
      */
      var beat_matrix,
        _this = this;
      beat_matrix = [];
      _.each(_.range(1, msrs + 1), function(m) {
        var i;
        i = m - 1;
        if (!beat_matrix[i]) beat_matrix[i] = [];
        return _.each(_.range(1, res + 1), function(b) {
          var this_beat;
          if (app.util.compare_nested_array(active_beats, m, b)) {
            this_beat = [m, b, 'active'];
          } else {
            this_beat = [m, b, 'rest'];
          }
          return beat_matrix[i].push(this_beat);
        });
      });
      return beat_matrix;
    },
    draw_measure: function(w, h, margin, a, d, s, r) {
      /*
      
      		Returns a function that takes one row from the beat_matrix (`msr`) and returns
      		a set of drawing instructions for that measure.
      */      return function(msr) {
        var active_b, end, instructions, rest_b;
        active_b = app.beats.amp_env(w, h, margin, a, d, s, r);
        rest_b = app.beats.rest_beat(w, h, margin);
        instructions = [['beginPath'], ['moveTo', 0, h + (h + margin) * (msr[0][0] - 1)]];
        end = [['closePath'], ['stroke'], ['fill']];
        _.each(msr, function(beat) {
          var b, instrs, m, type;
          m = beat[0], b = beat[1], type = beat[2];
          instrs = type === 'active' ? active_b(m, b) : rest_b(m, b);
          return _.each(instrs, function(instr) {
            return instructions.push(instr);
          });
        });
        _.each(end, function(instr) {
          return instructions.push(instr);
        });
        return instructions;
      };
    },
    get_instrs: function(m, res, a, d, s, r, active_beats) {
      /*
      
      		Produces a beat matrix, and maps `draw_measure` against it, returning a full
      		set of beatmap drawing instructions. Height and margin defaults are set here.
      */
      var beat_info, beat_matrix, drawer, h, margin, msr, w, _i, _len, _results;
      beat_info = app.beats.get_adjusted_beat_positions(res, active_beats);
      w = 1024 / res;
      h = 100;
      margin = 10;
      drawer = app.beats.draw_measure(w, h, margin, a, d, s, r);
      beat_matrix = app.beats.get_beat_matrix(res, m, beat_info);
      _results = [];
      for (_i = 0, _len = beat_matrix.length; _i < _len; _i++) {
        msr = beat_matrix[_i];
        _results.push(drawer(msr));
      }
      return _results;
    }
  };

}).call(this);
