(function() {
  var app;

  if (!window.coffeetable) window.coffeetable = {};

  app = window.coffeetable;

  app.util = {
    knobify: function(el, min, max, val, on_change) {
      /*
      
      		Takes an element (to be converted into a knob), minimum, maximum, and default
      		values, and a function to call on change. The on_change function is delayed 250ms
      		(see DelayedWithInterrupt, above).
      */
      var deferer;
      deferer = app.util.delay_with_interrupt(on_change, 250);
      $(el).attr('value', val);
      return $(el).knob({
        'width': 90,
        'min': min,
        'max': max,
        'change': function(v) {
          return deferer.start_timer(v);
        }
      });
    },
    delay_with_interrupt: function(func, ms_delay) {
      /*
      
      		Takes a function and the number of milliseconds to wait before executing it.
      		Returns an object with `start_timer` and `cancel` methods; if `start_timer`
      		is called again before `ms_delay` has transipired, the timer is reset.
      
      		This is useful for the BPM knob, which needs to have a little slack in it's
      		on_change function so that it isn't called too many times.
      */
      var deferer;
      return deferer = {
        start_timer: function(params) {
          deferer.cancel();
          return deferer.timeoutID = setTimeout(func, ms_delay, params);
        },
        cancel: function() {
          if (typeof deferer.timeoutID === 'number') {
            clearTimeout(deferer.timeoutID);
            return delete deferer.timeoutID;
          }
        }
      };
    },
    cycle: function(arr) {
      var cycle;
      return cycle = {
        i: 0,
        len: arr.length,
        drop: function(n) {
          /*
          
          				Moves the index pointer `cycle.i` forward by `n`, looping back around to the beginning 
          				if `n` is greater than the length of the array being cycled.
          */
          var i;
          i = n + cycle.i;
          if (i < cycle.len) {
            if (i > -1) {
              cycle.i = i;
            } else {
              cycle.i = cycle.len - 1;
            }
          } else {
            cycle.i = i % cycle.len;
          }
          return this;
        },
        skip_to: function(target) {
          /*
          
          				Moves `cycle.i` to the index of `target` if `target` is in `arr`; returns `this`.
          */
          var new_i;
          new_i = _.indexOf(arr, target);
          if (new_i > -1) {
            cycle.i = new_i;
          } else {
            console.log("Error: target is not in this array.");
          }
          return this;
        },
        take: function(n, skip) {
          /*
          
          				Takes `n` elements from the cycled array, starting from wherever `cycle.i` has been set
          				and looping back to the start of the array if necessary. 
          				Either drops `skip` (if present) or resets `cycle.i` to `0` after completion.
          */
          var index, j, results;
          j = 0;
          results = [];
          while (j < n) {
            index = cycle.i + j;
            if (index >= cycle.len) index = index % cycle.len;
            j += 1;
            results.push(arr[index]);
          }
          if (skip) {
            cycle.drop(skip);
          } else {
            cycle.i = 0;
          }
          return results;
        },
        next: function() {
          cycle.i += 1;
          cycle.i = cycle.i === cycle.len ? 0 : cycle.i;
          return arr[cycle.i];
        },
        prev: function() {
          cycle.i -= 1;
          cycle.i = cycle.i < 0 ? cycle.len - 1 : cycle.i;
          return arr[cycle.i];
        },
        get_arr: function() {
          return arr;
        }
      };
    },
    compare_nested_array: function(active_beats, m, b) {
      /*
      
      		Because, according to Javascript, `[[1, 1]][0] !== [1, 1]`.
      */
      var in_array;
      in_array = false;
      _.each(active_beats, function(beat) {
        var ab, am;
        am = beat[0], ab = beat[1];
        if (am === m && ab === b) return in_array = true;
      });
      return in_array;
    }
  };

}).call(this);
